from ofxstatement.parser import CsvStatementParser
import csv

from . import TRANSACTION_TYPES

# The bank for some reason has varied the column labeling a lot.
# By collecting first lines of known CSV exports as signatures
# we catch any future changes to the CSV export format.

SIGNATURES = (
    "Arvop\xe4iv\xe4\tM\xe4\xe4r\xe4\Tapahtuma\tTilinumero\t"
    "Saaja/Maksaja\tViite\tViesti",
    "Kirjauspäivä	Arvopäivä	Maksupäivä	Määrä	Saaja/Maksaja	"
    "Tilinumero	BIC	Tapahtuma	Viite	Maksajan viite	Viesti	"
    "Kortinnumero	Kuitti",
)

COLUMNS = 12


class NordeaCsvStatementParser(CsvStatementParser):
    """parser for various variations with common field semantics"""

    mappings = {
        "date": 0,
        "amount": 3,
        "payee": 4,
        "trntype": 7,
        "refnum": 8,
        "memo": 9
    }

    date_format = "%d.%m.%Y"

    def split_records(self):
        return csv.reader(self.fin, delimiter='\t', quotechar='"')

    def parse_record(self, line):

        # Change decimalsign from , to .
        line[3] = line[3].replace(',', '.')

        # Set transaction type
        transaction_type = line[7].upper()
        if transaction_type in TRANSACTION_TYPES:
            line[7] = TRANSACTION_TYPES[transaction_type]
        else:
            line[7] = 'DEBIT' if line[3][0] == '-' else 'CREDIT'

        # the CSV leaves off empty data at the end of record
        # so we fill in some blanks
        missing = COLUMNS - len(line)
        line.extend(missing * [""])

        # fill statement line according to mappings
        sl = super(NordeaCsvStatementParser, self).parse_record(line)
        return sl



from ofxstatement.plugin import Plugin
from ofxstatement.parser import StatementParser
from ofxstatement.statement import Statement, StatementLine

import sys
from xml.etree import ElementTree
import hashlib



class PostbankXMLPlugin(Plugin):
    def get_parser(self, filename):
        encoding = self.settings.get('charset', 'iso-8859-1')
        file_obj = open(filename, 'r', encoding=encoding)
        parser = PostbankXMLParser(file_obj)
        parser.statement.account_id = self.settings['account']
        #parser.statement.bank_id = self.settings['bank']
        parser.statement.currency = self.settings.get('currency', 'EUR')
        return parser


class PostbankXMLParser(StatementParser):
    date_format = "%Y-%m-%d"

    def __init__(self, file_obj):
        super(PostbankXMLParser,self).__init__()
        self.__data = file_obj.read()
        self.statement = Statement()


    def P(self,*names):
        res = []
        for name in names:
            res.append("{urn:iso:std:iso:20022:tech:xsd:camt.052.001.03}" + name)
        return '/'.join(res)

    def split_records(self):
        root = ElementTree.fromstring(self.__data)
        return root.findall(self.P('BkToCstmrAcctRpt','Rpt','Ntry'))

    def parse_record(self, record):
        P = self.P

        sl = StatementLine()

        sl.date = self.parse_datetime(record.findall(P('ValDt','Dt'))[0].text)
        sl.amount = float(record.findall(P('Amt'))[0].text)
        sl.trntype = 'DEBIT' if (record.findall(P('CdtDbtInd'))[0].text == "DBIT") else 'CREDIT'
        if sl.trntype == 'DEBIT' and sl.amount > 0:
            sl.amount = -sl.amount
        sl.payee = record.findall(P('NtryDtls','TxDtls','RltdPties','Cdtr','Nm'))[0].text
        sl.memo = ' '.join([e.text for e in record.findall(P('NtryDtls','TxDtls','RmtInf','Ustrd'))])

        # generate unique id
        h = hashlib.sha256()
        h.update(str(sl.date).encode('utf-8'))
        h.update(str(sl.amount).encode('utf-8'))
        h.update(str(sl.trntype).encode('utf-8'))
        h.update(str(sl.payee).encode('utf-8'))
        h.update(str(sl.memo).encode('utf-8'))

        sl.id = h.hexdigest()

        return sl






import csv

from ofxstatement.plugin import Plugin
from ofxstatement.parser import CsvStatementParser
from ofxstatement.statement import Statement, StatementLine, BankAccount, recalculate_balance
from pdb import set_trace as bp


class TriodosDePlugin(Plugin):
    """German Triodos plugin for ofxstatement
    """

    def get_parser(self, filename):
        f = open(filename, 'r', encoding=self.settings.get("charset", "ISO-8859-1"))
        parser = TriodosDeParser(f)
        return parser


class TriodosDeParser(CsvStatementParser):
    date_format = "%d.%m.%Y"
    mappings = {
        'date': 0,
        'payee': 3,
        'memo': 8,
        'amount': 11
    }

    def __init__(self, filename):
        self.statement = Statement('TRIODEF1', None,'EUR')
        self.fin = filename
        self.statement.currency = 'EUR'

    def parse(self):
        """Main entry point for parsers

        super() implementation will call to split_records and parse_record to
        process the file.
        """
        stmt = super(TriodosDeParser, self).parse()
        recalculate_balance(stmt)
        return stmt

    def split_records(self):
        """Return iterable object consisting of a line per transaction
        """
        return csv.reader(self.fin, delimiter=';')

    def parse_record(self, line):
        """Parse given transaction line and return StatementLine object
        """
        stmtline = super(TriodosDeParser, self).parse_record(line)
        """ As the amount is always positive, we negate it,
        if the amount is debit (= S[oll])
        """
        if line[12] == "S":
            stmtline.amount = -stmtline.amount
        return stmtline

    def parse_float(self, value):
        """Return a float from a string with ',' as decimal mark.
         ex. 1.234,56 -> 1234.56
        """
        return float(value.replace('.','').replace(',', '.'))



import xlrd
from datetime import datetime
from ofxstatement import statement
from ofxstatement.plugin import Plugin
from ofxstatement.parser import StatementParser


class FinecoPlugin(Plugin):
    """italian bank Fineco, it parses both xls files available for private accounts
    """

    def get_parser(self, filename):
        return FinecoStatementParser(filename)


class FinecoStatementParser(StatementParser):

    # HomeBank import payee/<NAME> in its description field, that can be used by assignement rules
    memo2payee = True
    date_format = '%d/%m/%Y'
    bank_id = 'FinecoBank'
    currency = 'EUR'
    tpl = {
        'savings' : {
            'th' : [
                u"Data",
                u"Entrate",
                u"Uscite",
                u"Descrizione",
                u"Descrizione Completa",
                u"Stato",
            ],
            'account_id_pos' : [1, 0],
            'account_id_str' : 'Conto Corrente: ',
            'xfer_str' : 'Bonifico ',
            'cash_str' : 'Prelievi Bancomat ',
            'extra_field' : 'Moneymap',
        },
        # this will be dropped as soon as it will not be available to download
        'savings_legacy' : {
            'th' : [
                u"Data Operazione",
                u"Data Valuta",
                u"Entrate",
                u"Uscite",
                u"Descrizione",
                u"Descrizione Completa",
            ],
            'account_id_pos' : [0, 0],
            'account_id_str' : 'Conto Corrente: ',
            'xfer_str' : 'Bonifico ',
            'cash_str' : 'Prelievi Bancomat ',
            'extra_field' : 'Money Map',
        },
        'cards' : {
            'th' : [
                u"Data operazione",
                u"Data registrazione",
                u"Descrizione",
                u"Tipo spesa",
                u"Tipo rimborso",
                u"Importo in EUR",
            ],
            'th_row' : 0,
            'amount_field' : 5,
            'account_id_pos' : [1, 2],
            'account_id_str' : ' **** **** ',
        }
    }
    common_footer_marker = 'Totale'
    th_separator_idx = 0
    cur_tpl = 'savings'
    extra_field = False;


    def __init__(self, filename):
        self.filename = filename


    def parse(self):
        """Main entry point for parsers

        super() implementation will call to split_records and parse_record to
        process the file.
        """
        workbook = xlrd.open_workbook(self.filename)
        sheet = workbook.sheet_by_index(0)
        heading, rows = [], []

        for rowidx in range(sheet.nrows):
            row = sheet.row_values(rowidx)

            # issue #5 and #3: dates might be formatted as excel dates (floats) rather than strings
            if type(row[0]) is float:
                row[0] = datetime.strftime(xlrd.xldate_as_datetime(row[0], 0), self.date_format)

            # split heading from current statement
            if self.th_separator_idx > 0:
                if row[0] != '' and not row[0].startswith(self.common_footer_marker):
                    rows.append(row)
            else:
                heading.append(row)

            # guess sheet tpl type
            for name, tpl in self.tpl.items():
                if row[0] == tpl['th'][0]:
                    self.th_separator_idx = rowidx
                    self.cur_tpl = name

        # issue #1: check if the file has the "Money Map" extra field
        if 'extra_field' in self.tpl[self.cur_tpl] and heading[-1][-1] == self.tpl[self.cur_tpl]['extra_field']:
            self.tpl[self.cur_tpl]['th'].append(self.tpl[self.cur_tpl]['extra_field'])
            self.extra_field = True

        # issue #2: some cards statements could miss "Tipo Spesa" and "Tipo Rimborso" columns
        elif self.cur_tpl == 'cards' and heading[-1][3] == self.tpl['cards']['th'][5]:
            self.tpl['cards']['th'].remove("Tipo spesa")
            self.tpl['cards']['th'].remove("Tipo rimborso")
            self.tpl['cards']['th'].append("")
            self.tpl['cards']['th_row'] = 1
            self.tpl['cards']['amount_field'] = 4

        self.validate(heading)

        row = self.tpl[self.cur_tpl]['account_id_pos'][0]
        col = self.tpl[self.cur_tpl]['account_id_pos'][1]
        account_id = sheet.cell_value(row, col).replace(self.tpl[self.cur_tpl]['account_id_str'], '')

        self.statement = statement.Statement(
            bank_id = self.bank_id,
            account_id = account_id,
            currency = self.currency
        )

        self.rows = rows
        return super(FinecoStatementParser, self).parse()


    def validate(self, heading):
        if self.th_separator_idx == 0:
            raise ValueError('unkown file')

        current_header = heading[self.th_separator_idx]
        msg = None

        row = self.tpl[self.cur_tpl]['account_id_pos'][0]
        col = self.tpl[self.cur_tpl]['account_id_pos'][1]
        if self.cur_tpl != 'cards' and not heading[row][col].startswith(self.tpl[self.cur_tpl]['account_id_str']):
            msg = "No account id cell found"

        elif self.tpl[self.cur_tpl]['th'] != current_header:
            msg = "\n".join([
                "Header template doesn't match:",
                "expected: %s" % self.tpl[self.cur_tpl]['th'],
                "current  : %s" % current_header
            ])

        if msg:
            raise ValueError(msg)


    # returns a negative number as outcome or a positive one as income
    def calc_amount(self, income, outcome):
        assert (income > 0) ^ (outcome != 0)
        if income > 0:
            return income
        elif outcome != 0:
            return -1 * outcome
        else:
            return 0.0


    def split_records(self):
        """Return iterable object consisting of a line per transaction
        """
        for row in self.rows:
            yield row


    def parse_record(self, row):
        """Parse given transaction line and return StatementLine object
        """
        stmt_line = statement.StatementLine()

        if self.cur_tpl == 'savings' or self.cur_tpl == 'savings_legacy':
            col_shift = 1 if self.cur_tpl == 'savings_legacy' else 0
            if row[1+col_shift]:
                income = row[1+col_shift]
                outcome = 0
                stmt_line.trntype = "CREDIT"
            elif row[2+col_shift]:
                outcome = row[2+col_shift]
                income = 0
                stmt_line.trntype = "DEBIT"

            memo_short = row[3+col_shift]
            if memo_short.startswith(self.tpl['savings']['xfer_str']):
                stmt_line.trntype = "XFER"
            elif memo_short.startswith(self.tpl['savings']['cash_str']):
                stmt_line.trntype = "CASH"

            stmt_line.memo = row[4+col_shift]
            if self.extra_field and row[6] != '':
                stmt_line.memo = stmt_line.memo + ' - ' + row[6]

            stmt_line.amount = self.calc_amount(income, outcome)

        elif self.cur_tpl == 'cards':
            if row[3] == 'P':
                stmt_line.trntype = "CASH"

            if row[ self.tpl['cards']['amount_field'] ] < 0:
                stmt_line.trntype = "DEBIT"
            else:
                stmt_line.trntype = "CREDIT"

            stmt_line.memo = row[2]
            stmt_line.amount = row[ self.tpl['cards']['amount_field'] ]

        if self.memo2payee:
            stmt_line.payee = stmt_line.memo

        stmt_line.date = datetime.strptime(row[0], self.date_format)
        stmt_line.id = statement.generate_transaction_id(stmt_line)

        return stmt_line



from typing import Optional, List
import xml.etree.ElementTree as ET
import datetime
import re
from decimal import Decimal

from ofxstatement import exceptions
from ofxstatement.plugin import Plugin
from ofxstatement.statement import Statement, StatementLine
from ofxstatement.parser import AbstractStatementParser


ISO20022_NAMESPACE_ROOT = "urn:iso:std:iso:20022:tech:xsd:camt.053.001"

CD_CREDIT = "CRDT"
CD_DEBIT = "DBIT"


class Iso20022Plugin(Plugin):
    """ISO-20022 plugin"""

    def get_parser(self, filename: str) -> "Iso20022Parser":
        default_ccy = self.settings.get("currency")
        parser = Iso20022Parser(filename, currency=default_ccy)
        return parser


class Iso20022Parser(AbstractStatementParser):
    def __init__(self, filename: str, currency: str = None):
        self.filename = filename
        self.currency = currency

    def parse(self) -> Statement:
        """Main entry point for parsers"""
        self.statement = Statement()
        self.statement.currency = self.currency
        tree = ET.parse(self.filename)

        # Find out XML namespace and make sure we can parse it
        ns = self._get_namespace(tree.getroot())
        if not ns.startswith(ISO20022_NAMESPACE_ROOT):
            raise exceptions.ParseError(0, "Cannot recognize ISO20022 XML")

        self.xmlns = {"s": ns}

        self._parse_statement_properties(tree)
        self._parse_lines(tree)

        return self.statement

    def _get_namespace(self, elem: ET.Element) -> str:
        m = re.match(r"\{(.*)\}", elem.tag)
        return m.groups()[0] if m else ""

    def _parse_statement_properties(self, tree: ET.ElementTree) -> None:
        stmt = tree.find("./s:BkToCstmrStmt/s:Stmt", self.xmlns)
        assert stmt is not None

        bnk = stmt.find("./s:Acct/s:Svcr/s:FinInstnId/s:BIC", self.xmlns)
        if bnk is None:
            bnk = stmt.find("./s:Acct/s:Svcr/s:FinInstnId/s:Nm", self.xmlns)
        iban = stmt.find("./s:Acct/s:Id/s:IBAN", self.xmlns)
        assert iban is not None
        ccy = stmt.find("./s:Acct/s:Ccy", self.xmlns)
        bals = stmt.findall("./s:Bal", self.xmlns)

        acctCurrency = ccy.text if ccy is not None else None
        if acctCurrency:
            self.statement.currency = acctCurrency
        else:
            if self.statement.currency is None:
                raise exceptions.ParseError(
                    0,
                    "No account currency provided in statement. Please "
                    "specify one in configuration file (e.g. currency=EUR)",
                )

        bal_amts = {}
        bal_dates = {}
        for bal in bals:
            cd = bal.find("./s:Tp/s:CdOrPrtry/s:Cd", self.xmlns)
            amt = bal.find("./s:Amt", self.xmlns)
            dt = bal.find("./s:Dt", self.xmlns)
            assert cd is not None
            assert amt is not None
            amt_ccy = amt.get("Ccy")
            # Amount currency should match with statement currency
            if amt_ccy != self.statement.currency:
                continue

            bal_amts[cd.text] = self._parse_amount(amt)
            bal_dates[cd.text] = self._parse_date(dt)

        if not bal_amts:
            raise exceptions.ParseError(
                0,
                "No statement balance found for currency '%s'. Check "
                "currency of statement file." % self.statement.currency,
            )

        self.statement.bank_id = bnk.text if bnk is not None else None
        self.statement.account_id = iban.text
        self.statement.start_balance = bal_amts["OPBD"]
        self.statement.start_date = bal_dates["OPBD"]
        self.statement.end_balance = bal_amts["CLBD"]
        self.statement.end_date = bal_dates["CLBD"]

    def _parse_lines(self, tree: ET.ElementTree) -> None:
        stmt = tree.find("./s:BkToCstmrStmt/s:Stmt", self.xmlns)
        assert stmt is not None

        for ntry in self._findall(stmt, "Ntry"):
            sline = self._parse_line(ntry)
            if sline is not None:
                self.statement.lines.append(sline)

    def _parse_line(self, ntry: ET.Element) -> Optional[StatementLine]:
        sline = StatementLine()

        crdeb = self._findstrict(ntry, "CdtDbtInd").text

        amtnode = self._findstrict(ntry, "Amt")
        amt_ccy = amtnode.get("Ccy")

        if amt_ccy != self.statement.currency:
            # We can't include amounts with incompatible currencies into the
            # statement.
            return None

        amt = self._parse_amount(amtnode)
        if crdeb == CD_DEBIT:
            amt = -amt
            payee = self._find(ntry, "NtryDtls/TxDtls/RltdPties/Cdtr/Nm")
        else:
            payee = self._find(ntry, "NtryDtls/TxDtls/RltdPties/Dbtr/Nm")

        sline.payee = payee.text if payee is not None else None
        sline.amount = amt

        dt = self._find(ntry, "ValDt")
        sline.date = self._parse_date(dt)

        bookdt = self._find(ntry, "BookgDt")
        sline.date_user = self._parse_date(bookdt)

        svcref = self._find(ntry, "NtryDtls/TxDtls/Refs/AcctSvcrRef")
        if svcref is None:
            svcref = self._find(ntry, "AcctSvcrRef")
        if svcref is not None:
            sline.refnum = svcref.text

        # Try to find memo from different possible locations
        refinf = self._find(ntry, "NtryDtls/TxDtls/RmtInf/Strd/CdtrRefInf/Ref")
        rmtinf = self._find(ntry, "NtryDtls/TxDtls/RmtInf/Ustrd")
        addinf = self._find(ntry, "AddtlNtryInf")
        if refinf is not None:
            sline.memo = refinf.text
        elif rmtinf is not None:
            sline.memo = rmtinf.text
        elif addinf is not None:
            sline.memo = addinf.text

        return sline

    def _parse_date(self, dtnode: Optional[ET.Element]) -> Optional[datetime.datetime]:
        if dtnode is None:
            return None

        dt = self._find(dtnode, "Dt")
        dttm = self._find(dtnode, "DtTm")

        if dt is not None and dt.text is not None:
            dtvalue = self._notimezone(dt.text)
            return datetime.datetime.strptime(dtvalue, "%Y-%m-%d")
        else:
            assert dttm is not None and dttm.text is not None
            dtvalue = self._notimezone(dttm.text)
            return datetime.datetime.strptime(dtvalue, "%Y-%m-%dT%H:%M:%S")

    def _notimezone(self, dt: str) -> str:
        # Sometimes we are getting time with ridiculous timezone, like
        # "2017-04-01+02:00", which is unparseable by any python parsers. Strip
        # out such timezone for good.
        if "+" not in dt:
            return dt
        dt, tz = dt.split("+")
        return dt

    def _parse_amount(self, amtnode: ET.Element) -> Decimal:
        assert amtnode.text is not None
        return Decimal(amtnode.text)

    def _find(self, tree: ET.Element, spath: str) -> Optional[ET.Element]:
        return tree.find(_toxpath(spath), self.xmlns)

    def _findstrict(self, tree: ET.Element, spath: str) -> ET.Element:
        found = self._find(tree, spath)
        if found is None:
            raise exceptions.ParseError(0, f"{spath} is not found in {tree}")
        return found

    def _findall(self, tree: ET.Element, spath: str) -> List[ET.Element]:
        return tree.findall(_toxpath(spath), self.xmlns)


def _toxpath(spath: str) -> str:
    tags = spath.split("/")
    path = ["s:%s" % t for t in tags]
    xpath = "./%s" % "/".join(path)
    return xpath



"""Parser for LITAS-ESIS csv statement"""
from typing import Optional, List, Iterable
import csv
from decimal import Decimal

from ofxstatement.parser import CsvStatementParser
from ofxstatement.plugin import Plugin
from ofxstatement import statement

LINETIME_HEADER = "000"
LINETYPE_TRANSACTION = "010"
LINETYPE_SUMMARY = "020"

SUMMARY_START = "LikutisPR"
SUMMARY_END = "LikutisPB"


class LitasEsisCsvDialect(csv.Dialect):
    delimiter = "\t"
    quotechar = None
    escapechar = None
    doublequote = False
    skipinitialspace = False
    lineterminator = "\r\n"
    quoting = csv.QUOTE_NONE


class LitasEsisCsvStatementParser(CsvStatementParser):
    date_format = "%Y%m%d"
    mappings = {
        "date": 2,
        "amount": 4,
        "id": 10,
        "payee": 17,
        "memo": 13,
    }
    charset = "cp1257"

    def parse_decimal(self, value: str) -> Decimal:
        return Decimal(value) / 100

    def split_records(self) -> Iterable[List[str]]:
        return csv.reader(self.fin, dialect=LitasEsisCsvDialect)

    def parse_record(self, line: List[str]) -> Optional[statement.StatementLine]:
        # print(line)

        linetype = line[0]
        stmt = self.statement

        if linetype == LINETIME_HEADER:
            # Get basic account information
            stmt.currency = line[17]
            stmt.bank_id = line[3]
            stmt.account_id = line[16] + line[17]
            return None

        elif linetype == LINETYPE_TRANSACTION:
            # parse transaction line in standard fasion
            stmtline = super(LitasEsisCsvStatementParser, self).parse_record(line)
            if stmtline is None:
                return None
            if line[6] == "D":
                assert stmtline.amount is not None
                stmtline.amount = -stmtline.amount
            return stmtline

        elif linetype == LINETYPE_SUMMARY:
            summarytype = line[1]
            if summarytype == SUMMARY_START:
                stmt.start_balance = self.parse_decimal(line[4])
                stmt.start_date = self.parse_datetime(line[2])
            elif summarytype == SUMMARY_END:
                stmt.end_balance = self.parse_decimal(line[4])
                stmt.end_date = self.parse_datetime(line[2])
        return None

    def swap_payee_and_memo(self) -> None:
        payee, memo = self.mappings["payee"], self.mappings["memo"]
        self.mappings["payee"] = memo
        self.mappings["memo"] = payee


class LitasEsisPlugin(Plugin):
    """Standard Lithuanian LITAS-ESIS format"""

    def get_parser(self, fin: str) -> LitasEsisCsvStatementParser:
        encoding = self.settings.get("charset", "utf-8")
        f = open(fin, "r", encoding=encoding)
        parser = LitasEsisCsvStatementParser(f)
        if "swap-payee-and-memo" in self.settings:
            parser.swap_payee_and_memo()
        return parser



import csv
from datetime import datetime
import re

from ofxstatement import statement
from ofxstatement.parser import CsvStatementParser
from ofxstatement.plugin import Plugin
from ofxstatement.statement import Statement


class UniCreditCZ16Plugin(Plugin):
    """UniCredit Bank Czech Republic and Slovakia a.s. (UNICREDIT format till 2016)
    """

    def get_parser(self, filename):
        UniCreditCZ16Plugin.encoding = self.settings.get('charset', 'utf-8')
        f = open(filename, "r", encoding=UniCreditCZ16Plugin.encoding)
        parser = UniCreditCZ16Parser(f)
        parser.statement.currency = self.settings.get('currency', 'CZK')
        parser.statement.bank_id = self.settings.get('bank', 'BACXCZPP')
        parser.statement.account_id = self.settings.get('account', '')
        parser.statement.account_type = self.settings.get('account_type', 'CHECKING')
        parser.statement.trntype = "OTHER"
        return parser


class UniCreditCZ16Parser(CsvStatementParser):

    # The columns are:
    #  0 Účet
    #  1 Částka
    #  2 Měna
    #  3 Datum zaúčtování
    #  4 Valuta
    #  5 Banka
    #  6 Název banky
    #  7 Název banky
    #  8 Číslo účtu
    #  9 Název účtu
    # 10 Adresa
    # 11 Adresa
    # 12 Adresa
    # 13 Detaily transakce
    # 14 Detaily transakce
    # 15 Detaily transakce
    # 16 Detaily transakce
    # 17 Detaily transakce
    # 18 Detaily transakce
    # 19 Konstatní kód
    # 20 Variabilní kód
    # 21 Specifický kód
    # 22 Platební titul
    # 23 Reference

    mappings = {"date_user": 3,
                "date": 4,
                "memo": 13,
                "payee": 9,
                "amount": 1,
                "check_no": 20,
                "refnum": 23, }

    date_format = "%Y-%m-%d"


    def split_records(self):
        return csv.reader(self.fin, delimiter=';', quotechar='"')


    def parse_record(self, line):
        if self.cur_record <= 3:
            return None

        sl = super(UniCreditCZ16Parser, self).parse_record(line)
        sl.date_user = datetime.strptime(sl.date_user, self.date_format)

        sl.id = statement.generate_transaction_id(sl)

        if sl.amount == 0:
            return None

        if line[13].startswith("SRÁŽKOVÁ DAŇ"):
            sl.trntype = "DEBIT"
        if line[7].startswith("ÚROKY"):
            sl.trntype = "INT"
        if line[7].startswith("Poplatek za "):
            sl.trntype = "FEE"
        if line[7].startswith("VRÁCENÍ POPLATKU ZA "):
            sl.trntype = "FEE"

        # sl.payee is imported as "Description" in GnuCash
        # sl.memo is imported as "Notes" in GnuCash
        # When sl.payee is empty, GnuCash imports sl.memo to "Description" and keeps "Notes" empty
        sl.memo = sl.memo + line[14] + line[15] + line[16] + line[17] + line[18]

        if not (line[8]  == "" or line[8]  == " "):
            sl.memo = sl.memo + "|ÚČ: "  + line[8]  + "/" + line[5]

        if not (line[20] == "" or line[20] == " "):
            sl.memo = sl.memo + "|VS: "  + line[20]

        if not (line[19] == "" or line[19] == " "):
            sl.memo = sl.memo + "|KS: "  + line[19]

        if not (line[21] == "" or line[21] == " "):
            sl.memo = sl.memo + "|SS: " + line[21]

        return sl

    # Substitute decimal point "," with "."
    # Remove thousands separator and and other text in value field
    def parse_float(self, value):
        value = re.sub(",", ".", value)
        value = re.sub("[ a-zA-Z]", "", value)
        return super(UniCreditCZ16Parser, self).parse_float(value)



#    Avangard Bank (https://www.avangard.ru/) plugin for ofxstatement
#
#    Copyright 2013 Andrey Lebedev <andrey@lebedev.lt>
#    Copyright 2016 Alexander Gerasiov <gq@cs.msu.su>
#    Copyright 2020 Dmitry Pavlov <zeldigas@gmail.com>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License version 3 as
#    published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ofxstatement.parser import StatementParser
from ofxstatement.plugin import Plugin
from ofxstatement import statement
from datetime import datetime
import csv

# file format options
av_delimiter = ';'
av_time_format = '%d.%m.%Y %H:%M'
av_encoding = 'cp1251'
av_currency = 'RUB'
av_fieldnames = ['tr_time', 'debit', 'credit', 'type', 'op_time', 'card', 'currency_value',
                 'currency', 'MCC', 'description']
av_type_map = {
    u"Зачисление": 'CREDIT',
    u"Покупка": 'PAYMENT',
    u"Типовой платеж": 'PAYMENT',
    u"Возврат": 'CREDIT',
    u"Перечисление средств по поручению клиента.": 'XFER',
    u"Перевод с карты": 'XFER',
    u"Продажа клиенту валюты": 'XFER',
    u"Перечисление средств на вклад": 'XFER',
    u"Выплата суммы бонусов по карте": 'DIV',
    u"Перечисление процентов по вкладу": 'DIV',
    u"Внесение на картсчет.": 'DEP',
    u"Комиссия за осуществление трансграничной операции": 'FEE',
    u"Комиссия за конверсию по трансграничной операции": 'FEE',
    u"Комиссия за предоставление овердрафта": 'FEE',
    u"Комиссия за получение наличных": 'FEE',
    u"Комиссия за операцию": 'FEE',
    u"Комиссия за осуществление расчетов по операциям": 'SRVCHG',
    u"Погашение процентов по предоставленному овердрафту": 'FEE',
    u"Плата за прием и обработку платежных документов": 'FEE',
    u"Снятие со счета": 'CASH',
    u"Внесение на счет": 'DEP',
    u"Наличные": 'ATM',
    u"Погашение овердрафта": None,
    u"Предоставление овердрафта": None,
}


def parse_type(type, amount):
    for filter in av_type_map.keys():
        if type.startswith(filter):
            return av_type_map[filter]

    result = None

    if amount > 0:
        result = 'DEBIT'
    elif amount < 0:
        result = 'CREDIT'

    # print("Unknown type \"%s\", consider %s"%(type, result))
    return result


class AvangardStatementParser(StatementParser):
    statement = None

    def __init__(self, fin):
        super().__init__()
        self.statement = statement.Statement()
        self.fin = fin

    def split_records(self):
        return csv.DictReader(self.fin, delimiter=av_delimiter, fieldnames=av_fieldnames)

    def parse_record(self, line):
        transaction = statement.StatementLine()

        transaction.date = datetime.strptime(line[('op_time' if line['op_time'] else 'tr_time')],
                                             av_time_format)

        transaction.amount = (float(line['debit']) if line['debit'] else 0) - (
            float(line['credit']) if line['credit'] else 0)

        transaction.trntype = parse_type(line['type'], transaction.amount)

        transaction.memo = line['description'] if line['description'] else line['type']

        if line['MCC']:
            transaction.memo = "%s, %s" % (transaction.memo, line['MCC'])

        if line['card']:
            transaction.memo = "%s, %s" % (transaction.memo, line['card'])

        # as csv file does not contain explicit id of transaction, generating artificial one
        transaction.id = statement.generate_transaction_id(transaction)

        if transaction.trntype:
            return transaction
        else:
            return None


class AvangardPlugin(Plugin):
    """Avangard Bank CSV (http://avangard.ru)
    """

    def get_parser(self, fin):
        f = open(fin, 'r', encoding=av_encoding)
        parser = AvangardStatementParser(f)
        parser.statement.currency = self.settings.get('currency', av_currency)
        parser.statement.account_id = self.settings['account']
        parser.statement.bank_id = self.settings.get('bank', 'Avangard')
        return parser



#!/usr/bin/env python3
# This file is part of ofxstatement-austrian.
# See README.rst for more information.

import datetime
from decimal import Decimal
import os
import unittest
from ofxstatement.statement import generate_transaction_id

from ofxstatement.plugins.ingdiba import IngDiBaCsvParser


class TestLivebankCsvParser(unittest.TestCase):
    """Unit tests for IngDiBaCsvParser."""

    def setUp(self):
        csvfile = os.path.join(
            os.path.dirname(__file__), 'samples', 'ing-diba.csv')
        with open(csvfile, 'r', encoding='iso-8859-1') as fin:
            self.statement = IngDiBaCsvParser(fin).parse()

    def test_statement_properties(self):
        self.assertEqual(len(self.statement.lines), 6)
        self.assertEqual(self.statement.start_balance, Decimal('0.0'))
        self.assertEqual(self.statement.end_balance, Decimal('-992.03'))
        self.assertEqual(self.statement.currency, "EUR")
        self.assertEqual(self.statement.account_id, "12345678001")
        self.assertEqual(
            self.statement.start_date, datetime.datetime(2013, 8, 13, 0, 0))
        self.assertEqual(
            self.statement.end_date, datetime.datetime(2013, 12, 31, 0, 0))

    def test_line0_interest_earned(self):
        line = self.statement.lines[0]
        self.assertEqual(line.amount, Decimal('12.23'))
        self.assertEqual(line.memo, "Habenzinsen")
        self.assertEqual(line.trntype, "CREDIT")
        self.assertEqual(line.date, datetime.datetime(2013, 12, 31, 0, 0))
        self.assertEqual(line.id, generate_transaction_id(line))

    def test_line1_interest_paid(self):
        line = self.statement.lines[1]
        self.assertEqual(line.amount, Decimal('-34.56'))
        self.assertEqual(line.memo, "Kapitalertragsteuer")
        self.assertEqual(line.trntype, "DEBIT")
        self.assertEqual(line.date, datetime.datetime(2013, 12, 31, 0, 0))
        self.assertEqual(line.id, generate_transaction_id(line))

    def test_line2_credit(self):
        line = self.statement.lines[2]
        self.assertEqual(line.amount, Decimal('500.00'))
        self.assertEqual(line.memo, "Eingang: XXX")
        self.assertEqual(line.trntype, "CREDIT")
        self.assertEqual(line.date, datetime.datetime(2013, 12, 23, 0, 0))
        self.assertEqual(line.id, generate_transaction_id(line))

    def test_line3_transfer_credit(self):
        line = self.statement.lines[3]
        self.assertEqual(line.amount, Decimal('20.30'))
        self.assertEqual(line.memo, "Umbuchung von 003")
        self.assertEqual(line.trntype, "CREDIT")
        self.assertEqual(line.date, datetime.datetime(2013, 9, 25, 0, 0))
        self.assertEqual(line.id, generate_transaction_id(line))

    def test_line4_debit(self):
        line = self.statement.lines[4]
        self.assertEqual(line.amount, Decimal('-1500'))
        self.assertEqual(line.memo, "Auszahlung - XXX")
        self.assertEqual(line.trntype, "DEBIT")
        self.assertEqual(line.date, datetime.datetime(2013, 8, 27, 0, 0))
        self.assertEqual(line.id, generate_transaction_id(line))

    def test_line5_bonus(self):
        line = self.statement.lines[5]
        self.assertEqual(line.amount, Decimal('10.00'))
        self.assertEqual(line.memo, "Prämie Foo")
        self.assertEqual(line.trntype, "CREDIT")
        self.assertEqual(line.date, datetime.datetime(2013, 8, 13, 0, 0))
        self.assertEqual(line.id, generate_transaction_id(line))

# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 smartindent autoindent



import datetime
from decimal import Decimal
from unittest import mock

from ofxstatement.plugins.vtb import VtbPlugin
from .util import file_sample


def test_parser():
    plugin = VtbPlugin(mock.Mock(), {'currency': 'RUR'})
    statement = plugin.get_parser(file_sample("vtb.csv")).parse()
    assert statement.bank_id == 'VTB'
    assert statement.currency == 'RUR'
    assert statement.account_id == '462235******0069'
    assert statement.start_date == datetime.datetime(2019, 7, 7, 0, 0)
    assert statement.end_date == datetime.datetime(2019, 7, 13, 0, 0)
    assert statement.start_balance == Decimal('99955.01')
    assert statement.end_balance == Decimal('82604.01')
    assert len(statement.lines) == 4
    _check_line(statement.lines[0], '-336.15', datetime.datetime(2019, 7, 8, 0, 0), 'Карта *1234 payee1', 'payee1')
    _check_line(statement.lines[1], '-283.48', datetime.datetime(2019, 7, 7, 0, 0), 'Карта *1234 payee2', 'payee2')
    _check_line(statement.lines[2], '-300.00', datetime.datetime(2019, 7, 7, 0, 0), 'Карта *1234 PAY MTS RU TOPUP 1234',
                'PAY MTS RU TOPUP 1234')
    _check_line(statement.lines[3], '-600.00', None, 'Карта *1234 MEGAFON TOPUP  1234', 'MEGAFON TOPUP  1234')


def test_parser_with_user_dates():
    plugin = VtbPlugin(mock.Mock(), {'currency': 'RUR', 'user_date': 'true'})
    parser = plugin.get_parser(file_sample("vtb_user_date.csv"))
    assert parser.user_date
    _check_line(parser.parse().lines[0], '-336.15', datetime.datetime(2019, 7, 6, 0, 0), 'Карта *1234 payee1', 'payee1')


def _check_line(line, amount, date, memo, payee):
    assert line.amount == Decimal(amount)
    assert line.date == date
    assert line.memo == memo
    assert line.payee == payee



import csv
from datetime import datetime

from ofxstatement import statement
from ofxstatement.parser import CsvStatementParser
from ofxstatement.plugin import Plugin
#from ofxstatement.statement import Statement


class UniCreditCZPlugin(Plugin):
    """UniCredit Bank Czech Republic and Slovakia a.s. (CSV format since cca 2018, new banking)
    """

    def get_parser(self, filename):
        UniCreditCZPlugin.encoding = self.settings.get('charset', 'utf-8')
        f = open(filename, "r", encoding=UniCreditCZPlugin.encoding)
        parser = UniCreditCZParser(f)
        parser.statement.currency = self.settings.get('currency', 'CZK')
        parser.statement.bank_id = self.settings.get('bank', 'BACXCZPP')
        parser.statement.account_id = self.settings.get('account', '')
        parser.statement.account_type = self.settings.get('account_type', 'CHECKING')
        parser.statement.trntype = "XFER"
        return parser


class UniCreditCZParser(CsvStatementParser):

    # GnuCash recognises the following descriptive fields:
    # - Header for Transaction Journal:
    #   - Description
    #   - Notes
    # - Line item memo for each account in a single Transaction Journal:
    #   - Memo
    #
    # .payee is assigned to "Description" in GnuCash
    # .memo is assigned to "Memo" in GnuCash and also concatenated
    #       to "Notes" after "OFX ext. info" and "Trans type"
    #
    # When .payee is empty, GnuCash assigns .memo to:
    # - "Description" and does not concatenate to "Notes"
    # - "Memo"
    #
    # Although ofxstatement can create bank_account_to, GnuCash ignores it.
    #
    # In GnuCash, .check_no (if empty, then .refnum) is assigned to "Num".
    #
    # The approach is:
    # - merge counterparty name (.payee) + account number + bank code
    # - merge pmt reference (.memo) + other payment specifics (VS, KS, SS)

    # The columns are:
    #  0 Číslo účtu
    #  1 Částka
    #  2 Měna
    #  3 Datum rezervace
    #  4 Datum
    #  5 Kód banky protistrany : .payee 6
    #  6 Název banky protistrany 1 :
    #  7 Banka příjemce 2 :
    #  8 Číslo účtu protistrany : .payee 5
    #  9 Příjemce : .payee 1
    #  10 Adresa 1 : .payee 2
    #  11 Adresa 2 : .payee 3
    #  12 Adresa 3 : .payee 4
    #  13 Detaily transakce 1 : .memo 1
    #  14 Detaily transakce 2 : .memo 2
    #  15 Detaily transakce 3 : .memo 3
    #  16 Detaily transakce 4 : .memo 4
    #  17 Detaily transakce 5 : .memo 5
    #  18 Detaily transakce 6 : .memo 6
    #  19 KS : .memo 8
    #  20 VS : .memo 7 and .check_no
    #  21 SS : .memo 9
    #  22 Směnný kurz
    #  23 Referenční číslo
    #  24 Status
    #  25 Datum zamítnutí
    #  26 Detail zamítnutí
    #  27 Registrační číslo TPP
    #  28 Název TPP
    #  29 Reference příkazu TPP
    #  30 Země registrace
    #  31 Kód národní autority

    mappings = {
        "amount": 1,
        "date": 3,
        "date_user": 4,
        "payee": 9,
        "memo": 13,
        "check_no": 20,
        "refnum": 23,
    }

    #date_format = "%Y-%m-%d %H:%M:%S"
    date_format = "%Y-%m-%d"


    def split_records(self):
        return csv.reader(self.fin, delimiter=';', quotechar='"')


    def parse_record(self, line):
        # Ignore headers
        if self.cur_record <= 4:
            return None

        # Ignore incomplete lines
        if len(line) < 24:
            return None

        StatementLine = super(UniCreditCZParser, self).parse_record(line)

        StatementLine.id = statement.generate_transaction_id(StatementLine)

        if StatementLine.amount == 0:
            return None

        if   line[13].startswith("SRÁŽKOVÁ DAŇ"):
            StatementLine.trntype = "DEBIT"
        elif line[7].startswith("UROK DO"):
            StatementLine.trntype = "INT"
        elif line[7].startswith("Poplatek za "):
            StatementLine.trntype = "FEE"
        elif line[7].startswith("VRÁCENÍ POPLATKU ZA "):
            StatementLine.trntype = "FEE"

        StatementLine.memo = StatementLine.memo + " " + line[14]
        StatementLine.memo = StatementLine.memo + " " + line[15]
        StatementLine.memo = StatementLine.memo + " " + line[16]
        StatementLine.memo = StatementLine.memo + " " + line[17]
        StatementLine.memo = StatementLine.memo + " " + line[18]
        StatementLine.memo = StatementLine.memo.split()
        # Remove null items
        StatementLine.memo = [x for x in StatementLine.memo if x]
        StatementLine.memo = " ".join(StatementLine.memo)

        if not (line[20] == "" or line[20] == " "):
            StatementLine.memo = StatementLine.memo + "|VS:"  + line[20]

        if not (line[19] == "" or line[19] == " "):
            StatementLine.memo = StatementLine.memo + "|KS:"  + line[19]

        if not (line[21] == "" or line[21] == " "):
            StatementLine.memo = StatementLine.memo + "|SS:" + line[21]

        StatementLine.memo = StatementLine.memo.strip("|")

        # Counterparty name and address
        StatementLine.payee = StatementLine.payee + ", " + line[10].strip()
        StatementLine.payee = StatementLine.payee + ", " + line[11].strip()
        StatementLine.payee = StatementLine.payee + ", " + line[12].strip()
        StatementLine.payee = StatementLine.payee.split(", ")
        StatementLine.payee = [x for x in StatementLine.payee if x]
        StatementLine.payee = ", ".join(StatementLine.payee)

        # Bank account of counterparty
        StatementLine.payee = StatementLine.payee + "|" + line[8]
        StatementLine.payee = StatementLine.payee + "/" + line[5]
        StatementLine.payee = StatementLine.payee.strip("|")
        StatementLine.payee = StatementLine.payee.strip("/")

        return StatementLine

    # Remove thousands separator
    # Substitute decimal point "," with "."
    def parse_float(self, value):
        value = value.replace(".", "")
        value = value.replace(",", ".")
        return super(UniCreditCZParser, self).parse_float(value)



import csv
from datetime import datetime
import re

from ofxstatement import statement
from ofxstatement.parser import CsvStatementParser
from ofxstatement.plugin import Plugin


class AirBankCZPlugin(Plugin):
    """Air Bank a.s. (Czech Republic) (CSV, UTF-8)
    """

    def get_parser(self, filename):
        # Encoding may be set in Air Bank internet banking (Settings / Applications)
        AirBankCZPlugin.encoding = self.settings.get('charset', 'utf-8')
        f = open(filename, "r", encoding=AirBankCZPlugin.encoding)
        parser = AirBankCZParser(f)
        parser.statement.currency = self.settings.get('currency', 'CZK')
        parser.statement.bank_id = self.settings.get('bank', 'AIRACZPP')
        parser.statement.account_id = self.settings.get('account', '')
        parser.statement.account_type = self.settings.get('account_type', 'CHECKING')
        parser.statement.trntype = "OTHER"
        return parser


class AirBankCZParser(CsvStatementParser):
    date_format = "%d/%m/%Y"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, *kwargs)
        self.columns = None
        self.mappings = None

    def split_records(self):
        """Return iterable object consisting of a line per transaction
        """
        # Field delimiter may be set in Air Bank internet banking (Settings / Applications)
        return csv.reader(self.fin, delimiter=',', quotechar='"')

    def parse_record(self, line):
        """Parse given transaction line and return StatementLine object
        """

        # First line of CSV file contains headers, not an actual transaction
        if self.cur_record == 1:
            # Prepare columns headers lookup table for parsing
            self.columns = {v: i for i,v in enumerate(line)}
            self.mappings = {
                "date": self.columns['Datum provedení'],
                "memo": self.columns['Poznámka k úhradě'],
                "payee": self.columns['Název protistrany'],
                "amount": self.columns['Částka v měně účtu'],
                "check_no": self.columns['Variabilní symbol'],
                "refnum": self.columns['Referenční číslo'],
            }
            # And skip further processing by parser
            return None

        # Shortcut
        columns = self.columns

        # Normalize string
        for i,v in enumerate(line):
            line[i] = v.strip()

        if line[columns["Částka v měně účtu"]] == '':
            line[columns["Částka v měně účtu"]] = "0"
        if line[columns["Poplatek v měně účtu"]] == '':
            line[columns["Poplatek v měně účtu"]] = "0"

        StatementLine = super(AirBankCZParser, self).parse_record(line)

        # Ignore lines, which do not have posting date yet (typically pmts by debit cards
        # have some delays).
        if not line[columns["Datum zaúčtování"]]:
            return None
        else:
            StatementLine.date_user = line[columns["Datum zaúčtování"]]
            StatementLine.date_user = datetime.strptime(StatementLine.date_user, self.date_format)

        StatementLine.id = statement.generate_transaction_id(StatementLine)

        # Manually set some of the known transaction types
        payment_type = line[columns["Typ úhrady"]]
        if payment_type.startswith("Daň z úroku"):
            StatementLine.trntype = "DEBIT"
        elif payment_type.startswith("Kreditní úrok"):
            StatementLine.trntype = "INT"
        elif payment_type.startswith("Poplatek za "):
            StatementLine.trntype = "FEE"
        elif payment_type.startswith("Příchozí úhrada"):
            StatementLine.trntype = "XFER"
        elif payment_type.startswith("Odchozí úhrada"):
            StatementLine.trntype = "XFER"
        elif payment_type.startswith("Výběr hotovosti"):
            StatementLine.trntype = "ATM"
        elif payment_type.startswith("Platba kartou"):
            StatementLine.trntype = "POS"
        elif payment_type.startswith("Inkaso"):
            StatementLine.trntype = "DIRECTDEBIT"
        elif payment_type.startswith("Trvalý"):
            StatementLine.trntype = "REPEATPMT"
        else:
            print("WARN: Unexpected type of payment appeared - \"{}\". Using XFER transaction type instead".format(payment_type))
            StatementLine.trntype = "XFER"

        # .payee becomes OFX.NAME which becomes "Description" in GnuCash
        # .memo  becomes OFX.MEMO which becomes "Notes"       in GnuCash
        # When .payee is empty, GnuCash imports .memo to "Description" and keeps "Notes" empty

        # StatementLine.payee = "Název protistrany" + "Číslo účtu protistrany"
        if line[columns["Číslo účtu protistrany"]] != "":
            StatementLine.payee += "|ÚČ: " + line[columns["Číslo účtu protistrany"]]

        # StatementLine.memo = "Poznámka k úhradě" + the payment identifiers
        if line[columns["Variabilní symbol"]] != "":
            StatementLine.memo += "|VS: " + line[columns["Variabilní symbol"]]

        if line[columns["Konstantní symbol"]] != "":
            StatementLine.memo += "|KS: " + line[columns["Konstantní symbol"]]

        if line[columns["Specifický symbol"]] != "":
            StatementLine.memo += "|SS: " + line[columns["Specifický symbol"]]

        if line[columns["Název karty"]] != "":
            StatementLine.memo += "|Název karty: " + line[columns["Název karty"]]

        # Some type of fee is standalone, not related to transaction amount. Add it to amount field.only
        if float(line[columns["Poplatek v měně účtu"]]) != 0 and StatementLine.amount == 0:
            StatementLine.amount = float(line[columns["Poplatek v měně účtu"]])

        # Air Bank may show various fees on the same line as the underlying transaction.
        # In case there is a fee connected with the transaction, the fee is added as different transaction
        elif float(line[columns["Poplatek v měně účtu"]]) != 0 and StatementLine.amount != 0:
            fee_line = list(line)
            fee_line[columns['Částka v měně účtu']] = fee_line[columns["Poplatek v měně účtu"]]
            fee_line[columns['Poplatek v měně účtu']] = '0'
            fee_line[columns['Skupina plateb']] = "Poplatek za transakci"
            fee_line[columns["Poznámka k úhradě"]] = "Poplatek: " + fee_line[columns["Poznámka k úhradě"]]

            # parse the newly generated fee_line and append it to the rest of the statements
            stmt_line = self.parse_record(fee_line)
            if stmt_line:
                stmt_line.assert_valid()
                self.statement.lines.append(stmt_line)

        if StatementLine.amount == 0:
            return None

        return StatementLine



"""Parser for the statement of Poštovní spořitelna"""
from ofxstatement.parser import StatementParser
from ofxstatement.plugin import Plugin
from ofxstatement.statement import Statement, StatementLine
import re
import datetime
import logging

log = logging.getLogger('PS')

RecordParseRE = re.compile(r"\n{2,}", re.MULTILINE)


class PSTextFormatParser(StatementParser):
    # these keys mean in order: amount, date of transaction being recorded in
    # books, transaction type for statistical purposes, currency, name of
    # the other account, type of the transaction, memo, other account no.,
    # possible number of other sub-account, variable symbol (for identification
    # of the transaction), balance of the account after transaction
    #
    # note, that variable symbol is not guaranteed to be unique (it is quite
    # often number of the invoice or something which allows accountants to
    # identify the transaction on the account statement).
    _keys = set(('částka', 'datum zaúčtování', 'konstantní symbol', 'měna',
                 'název protiúčtu', 'označení operace', 'poznámka',
                 'protiúčet', 'specifický symbol', 'variabilní symbol',
                 'zůstatek',))

    statement = None
    currentLine = 0

    def __init__(self, fname):
        StatementParser.__init__(self)
        # bank_id=None, account_id=None, currency=None
        self.statement = Statement(currency="CZK")
        self.statement.start_date = datetime.date.fromtimestamp(0)
        self.statement.end_date = datetime.date.today()
        self.fin = fname

    def split_records(self):
        # uses self.fin
        return RecordParseRE.split(self.fin.read().strip())[1:]

    def parse_record(self, rec):
        MultilineRecRE = re.compile(r"\n\s+")
        SplitItemRE = re.compile(r":\s+")
        # join together multiline items
        rec = MultilineRecRE.sub(" ", rec)
        log.debug('rec = {}'.format(rec))
        res_iter = [SplitItemRE.split(line.strip(), maxsplit=1)
                    for line in rec.split("\n")]

        res_dict = dict((item for item in res_iter if len(item) > 1))
        res_key_set = set(res_dict.keys())
        log.debug('res_dict = {}'.format(res_dict))
        log.debug('res_key_set = {}'.format(res_key_set))

        # Insurance against unknown keys in the statement
        assert res_key_set <= self._keys, \
            "Unknown keys in the transaction: %s" % (res_key_set - self._keys)

        # no idea, how to make this unique and monotonically increasing
        try:
            trans_ID = res_dict['datum zaúčtování'] + res_dict['částka']
            if 'název protiúčtu' in res_dict:
                trans_ID += res_dict['název protiúčtu']
            if 'variabilní symbol' in res_dict:
                trans_ID += res_dict['variabilní symbol']
            date_str = res_dict['datum zaúčtování'].split('.')
        except KeyError:
            log.error('res_dict:\n{}'.format(res_dict))
            raise

        memo = ''
        if 'poznámka' in res_dict:
            memo = "%s\n" % res_dict['poznámka']
        if 'označení operace' in res_dict:
            memo += "%s\n" % res_dict['označení operace']
        stat_line = StatementLine(
            id=trans_ID,
            date=datetime.date(
                int(date_str[2]), int(date_str[1]), int(date_str[0])),
            memo=memo,
            amount=float(res_dict['částka']))

        if 'název protiúčtu' in res_dict:
            stat_line.payee = res_dict['název protiúčtu']

        # According to OFX spec this could be
        # "check (or other reference) no."
        # I don't see any requirement on monotonicity or uniqueness, but
        # I wonder how GnuCash understands that
        if 'variabilní symbol' in res_dict:
            stat_line.check_no = res_dict['variabilní symbol']

        return stat_line


class PSPlugin(Plugin):

    def get_parser(self, fin):
        encoding = self.settings.get('charset', 'utf-8-sig')
        inf = open(fin, encoding=encoding)
        return PSTextFormatParser(inf)
